# -*- coding: utf-8 -*- #

# Copyright 2015 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A module for dealing with unknown string and environment encodings."""

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

import sys

import six


def Encode(string, encoding=None):
    """Encode the text string to a byte string.

    Args:
      string: str, The text string to encode.
      encoding: The suggested encoding if known.

    Returns:
      str, The binary string.
    """
    if string is None:
        return None
    if not six.PY2:
        # In Python 3, the environment sets and gets accept and return text strings
        # only, and it handles the encoding itself so this is not necessary.
        return string
    if isinstance(string, six.binary_type):
        # Already an encoded byte string, we are done
        return string

    encoding = encoding or _GetEncoding()
    return string.encode(encoding)


def Decode(data, encoding=None):
    """Returns string with non-ascii characters decoded to UNICODE.

    UTF-8, the suggested encoding, and the usual suspects will be attempted in
    order.

    Args:
      data: A string or object that has str() and unicode() methods that may
        contain an encoding incompatible with the standard output encoding.
      encoding: The suggested encoding if known.

    Returns:
      A text string representing the decoded byte string.
    """
    if data is None:
        return None

    # First we are going to get the data object to be a text string.
    # Don't use six.string_types here because on Python 3 bytes is not considered
    # a string type and we want to include that.
    if isinstance(data, six.text_type) or isinstance(data, six.binary_type):
        string = data
    else:
        # Some non-string type of object.
        try:
            string = six.text_type(data)
        except (TypeError, UnicodeError):
            # The string cannot be converted to unicode -- default to str() which will
            # catch objects with special __str__ methods.
            string = str(data)

    if isinstance(string, six.text_type):
        # Our work is done here.
        return string

    try:
        # Just return the string if its pure ASCII.
        return string.decode("ascii")
    except UnicodeError:
        # The string is not ASCII encoded.
        pass

    # Try the suggested encoding if specified.
    if encoding:
        try:
            return string.decode(encoding)
        except UnicodeError:
            # Bad suggestion.
            pass

    # Try UTF-8 because the other encodings could be extended ASCII. It would
    # be exceptional if a valid extended ascii encoding with extended chars
    # were also a valid UITF-8 encoding.
    try:
        return string.decode("utf8")
    except UnicodeError:
        # Not a UTF-8 encoding.
        pass

    # Try the filesystem encoding.
    try:
        return string.decode(sys.getfilesystemencoding())
    except UnicodeError:
        # string is not encoded for filesystem paths.
        pass

    # Try the system default encoding.
    try:
        return string.decode(sys.getdefaultencoding())
    except UnicodeError:
        # string is not encoded using the default encoding.
        pass

    # We don't know the string encoding.
    # This works around a Python str.encode() "feature" that throws
    # an ASCII *decode* exception on str strings that contain 8th bit set
    # bytes. For example, this sequence throws an exception:
    #   string = '\xdc'  # iso-8859-1 'Ü'
    #   string = string.encode('ascii', 'backslashreplace')
    # even though 'backslashreplace' is documented to handle encoding
    # errors. We work around the problem by first decoding the str string
    # from an 8-bit encoding to unicode, selecting any 8-bit encoding that
    # uses all 256 bytes (such as ISO-8559-1):
    #   string = string.decode('iso-8859-1')
    # Using this produces a sequence that works:
    #   string = '\xdc'
    #   string = string.decode('iso-8859-1')
    #   string = string.encode('ascii', 'backslashreplace')
    return string.decode("iso-8859-1")


def GetEncodedValue(env, name, default=None):
    """Returns the decoded value of the env var name.

    Args:
      env: {str: str}, The env dict.
      name: str, The env var name.
      default: The value to return if name is not in env.

    Returns:
      The decoded value of the env var name.
    """
    name = Encode(name)
    value = env.get(name)
    if value is None:
        return default
    # In Python 3, the environment sets and gets accept and return text strings
    # only, and it handles the encoding itself so this is not necessary.
    return Decode(value)


def SetEncodedValue(env, name, value, encoding=None):
    """Sets the value of name in env to an encoded value.

    Args:
      env: {str: str}, The env dict.
      name: str, The env var name.
      value: str or unicode, The value for name. If None then name is removed from
        env.
      encoding: str, The encoding to use or None to try to infer it.
    """
    # Python 2 *and* 3 unicode support falls apart at filesystem/argv/environment
    # boundaries. The encoding used for filesystem paths and environment variable
    # names/values is under user control on most systems. With one of those values
    # in hand there is no way to tell exactly how the value was encoded. We get
    # some reasonable hints from sys.getfilesystemencoding() or
    # sys.getdefaultencoding() and use them to encode values that the receiving
    # process will have a chance at decoding. Leaving the values as unicode
    # strings will cause os module Unicode exceptions. What good is a language
    # unicode model when the module support could care less?
    name = Encode(name, encoding=encoding)
    if value is None:
        env.pop(name, None)
        return
    env[name] = Encode(value, encoding=encoding)


def EncodeEnv(env, encoding=None):
    """Encodes all the key value pairs in env in preparation for subprocess.

    Args:
      env: {str: str}, The environment you are going to pass to subprocess.
      encoding: str, The encoding to use or None to use the default.

    Returns:
      {bytes: bytes}, The environment to pass to subprocess.
    """
    encoding = encoding or _GetEncoding()
    return {Encode(k, encoding=encoding): Encode(v, encoding=encoding) for k, v in six.iteritems(env)}


def _GetEncoding():
    """Gets the default encoding to use."""
    return sys.getfilesystemencoding() or sys.getdefaultencoding()



# -*- coding: utf-8 -*- #
# Copyright 2013 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""General console printing utilities used by the Cloud SDK."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import signal
import subprocess
import sys

from afire.console import console_attr
from afire.console import console_pager
from afire.console import encoding
from afire.console import files


def IsInteractive(output=False, error=False, heuristic=False):
    """Determines if the current terminal session is interactive.

    sys.stdin must be a terminal input stream.

    Args:
      output: If True then sys.stdout must also be a terminal output stream.
      error: If True then sys.stderr must also be a terminal output stream.
      heuristic: If True then we also do some additional heuristics to check if
                 we are in an interactive context. Checking home path for example.

    Returns:
      True if the current terminal session is interactive.
    """
    if not sys.stdin.isatty():
        return False
    if output and not sys.stdout.isatty():
        return False
    if error and not sys.stderr.isatty():
        return False

    if heuristic:
        # Check the home path. Most startup scripts for example are executed by
        # users that don't have a home path set. Home is OS dependent though, so
        # check everything.
        # *NIX OS usually sets the HOME env variable. It is usually '/home/user',
        # but can also be '/root'. If it's just '/' we are most likely in an init
        # script.
        # Windows usually sets HOMEDRIVE and HOMEPATH. If they don't exist we are
        # probably being run from a task scheduler context. HOMEPATH can be '\'
        # when a user has a network mapped home directory.
        # Cygwin has it all! Both Windows and Linux. Checking both is perfect.
        home = os.getenv("HOME")
        homepath = os.getenv("HOMEPATH")
        if not homepath and (not home or home == "/"):
            return False
    return True


def More(contents, out, prompt=None, check_pager=True):
    """Run a user specified pager or fall back to the internal pager.

    Args:
      contents: The entire contents of the text lines to page.
      out: The output stream.
      prompt: The page break prompt.
      check_pager: Checks the PAGER env var and uses it if True.
    """
    if not IsInteractive(output=True):
        out.write(contents)
        return
    if check_pager:
        pager = encoding.GetEncodedValue(os.environ, "PAGER", None)
        if pager == "-":
            # Use the fallback Pager.
            pager = None
        elif not pager:
            # Search for a pager that handles ANSI escapes.
            for command in ("less", "pager"):
                if files.FindExecutableOnPath(command):
                    pager = command
                    break
        if pager:
            # If the pager is less(1) then instruct it to display raw ANSI escape
            # sequences to enable colors and font embellishments.
            less_orig = encoding.GetEncodedValue(os.environ, "LESS", None)
            less = "-R" + (less_orig or "")
            encoding.SetEncodedValue(os.environ, "LESS", less)
            # Ignore SIGINT while the pager is running.
            # We don't want to terminate the parent while the child is still alive.
            signal.signal(signal.SIGINT, signal.SIG_IGN)
            p = subprocess.Popen(pager, stdin=subprocess.PIPE, shell=True)
            enc = console_attr.GetConsoleAttr().GetEncoding()
            p.communicate(input=contents.encode(enc))
            p.wait()
            # Start using default signal handling for SIGINT again.
            signal.signal(signal.SIGINT, signal.SIG_DFL)
            if less_orig is None:
                encoding.SetEncodedValue(os.environ, "LESS", None)
            return
    # Fall back to the internal pager.
    console_pager.Pager(contents, out, prompt).Run()



# Copyright (C) 2018 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the decorators module."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from afire import core
from afire import decorators
from afire import testutils


class NoDefaults(object):
    """A class for testing decorated functions without default values."""

    @decorators.SetParseFns(count=int)
    def double(self, count):
        return 2 * count

    @decorators.SetParseFns(count=float)
    def triple(self, count):
        return 3 * count

    @decorators.SetParseFns(int)
    def quadruple(self, count):
        return 4 * count


@decorators.SetParseFns(int)
def double(count):
    return 2 * count


class WithDefaults(object):
    @decorators.SetParseFns(float)
    def example1(self, arg1=10):
        return arg1, type(arg1)

    @decorators.SetParseFns(arg1=float)
    def example2(self, arg1=10):
        return arg1, type(arg1)


class MixedArguments(object):
    @decorators.SetParseFns(float, arg2=str)
    def example3(self, arg1, arg2):
        return arg1, arg2


class PartialParseFn(object):
    @decorators.SetParseFns(arg1=str)
    def example4(self, arg1, arg2):
        return arg1, arg2

    @decorators.SetParseFns(arg2=str)
    def example5(self, arg1, arg2):
        return arg1, arg2


class WithKwargs(object):
    @decorators.SetParseFns(mode=str, count=int)
    def example6(self, **kwargs):
        return (
            kwargs.get("mode", "default"),
            kwargs.get("count", 0),
        )


class WithVarArgs(object):
    @decorators.SetParseFn(str)
    def example7(self, arg1, arg2=None, *varargs, **kwargs):  # pylint: disable=keyword-arg-before-vararg
        return arg1, arg2, varargs, kwargs


class FireDecoratorsTest(testutils.BaseTestCase):
    def testSetParseFnsNamedArgs(self):
        self.assertEqual(core.Fire(NoDefaults, command=["double", "2"]), 4)
        self.assertEqual(core.Fire(NoDefaults, command=["triple", "4"]), 12.0)

    def testSetParseFnsPositionalArgs(self):
        self.assertEqual(core.Fire(NoDefaults, command=["quadruple", "5"]), 20)

    def testSetParseFnsFnWithPositionalArgs(self):
        self.assertEqual(core.Fire(double, command=["5"]), 10)

    def testSetParseFnsDefaultsFromPython(self):
        # When called from Python, function should behave normally.
        self.assertTupleEqual(WithDefaults().example1(), (10, int))
        self.assertEqual(WithDefaults().example1(5), (5, int))
        self.assertEqual(WithDefaults().example1(12.0), (12, float))

    def testSetParseFnsDefaultsFromFire(self):
        # Fire should use the decorator to know how to parse string arguments.
        self.assertEqual(core.Fire(WithDefaults, command=["example1"]), (10, int))
        self.assertEqual(core.Fire(WithDefaults, command=["example1", "10"]), (10, float))
        self.assertEqual(core.Fire(WithDefaults, command=["example1", "13"]), (13, float))
        self.assertEqual(core.Fire(WithDefaults, command=["example1", "14.0"]), (14, float))

    def testSetParseFnsNamedDefaultsFromPython(self):
        # When called from Python, function should behave normally.
        self.assertTupleEqual(WithDefaults().example2(), (10, int))
        self.assertEqual(WithDefaults().example2(5), (5, int))
        self.assertEqual(WithDefaults().example2(12.0), (12, float))

    def testSetParseFnsNamedDefaultsFromFire(self):
        # Fire should use the decorator to know how to parse string arguments.
        self.assertEqual(core.Fire(WithDefaults, command=["example2"]), (10, int))
        self.assertEqual(core.Fire(WithDefaults, command=["example2", "10"]), (10, float))
        self.assertEqual(core.Fire(WithDefaults, command=["example2", "13"]), (13, float))
        self.assertEqual(core.Fire(WithDefaults, command=["example2", "14.0"]), (14, float))

    def testSetParseFnsPositionalAndNamed(self):
        self.assertEqual(core.Fire(MixedArguments, ["example3", "10", "10"]), (10, "10"))

    def testSetParseFnsOnlySomeTypes(self):
        self.assertEqual(core.Fire(PartialParseFn, command=["example4", "10", "10"]), ("10", 10))
        self.assertEqual(core.Fire(PartialParseFn, command=["example5", "10", "10"]), (10, "10"))

    def testSetParseFnsForKeywordArgs(self):
        self.assertEqual(core.Fire(WithKwargs, command=["example6"]), ("default", 0))
        self.assertEqual(core.Fire(WithKwargs, command=["example6", "--herring", '"red"']), ("default", 0))
        self.assertEqual(core.Fire(WithKwargs, command=["example6", "--mode", "train"]), ("train", 0))
        self.assertEqual(core.Fire(WithKwargs, command=["example6", "--mode", "3"]), ("3", 0))
        self.assertEqual(core.Fire(WithKwargs, command=["example6", "--mode", "-1", "--count", "10"]), ("-1", 10))
        self.assertEqual(core.Fire(WithKwargs, command=["example6", "--count", "-2"]), ("default", -2))

    def testSetParseFn(self):
        self.assertEqual(
            core.Fire(WithVarArgs, command=["example7", "1", "--arg2=2", "3", "4", "--kwarg=5"]),
            ("1", "2", ("3", "4"), {"kwarg": "5"}),
        )


if __name__ == "__main__":
    testutils.main()



# Copyright (C) 2018 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the helptext module."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import sys
import textwrap

from afire import formatting
from afire import helptext
from afire import test_components as tc
from afire import testutils
from afire import trace
import six


class HelpTest(testutils.BaseTestCase):
    def setUp(self):
        super(HelpTest, self).setUp()
        os.environ["ANSI_COLORS_DISABLED"] = "1"

    def testHelpTextNoDefaults(self):
        component = tc.NoDefaults
        help_screen = helptext.HelpText(component=component, trace=trace.FireTrace(component, name="NoDefaults"))
        self.assertIn("NAME\n    NoDefaults", help_screen)
        self.assertIn("SYNOPSIS\n    NoDefaults", help_screen)
        self.assertNotIn("DESCRIPTION", help_screen)
        self.assertNotIn("NOTES", help_screen)

    def testHelpTextNoDefaultsObject(self):
        component = tc.NoDefaults()
        help_screen = helptext.HelpText(component=component, trace=trace.FireTrace(component, name="NoDefaults"))
        self.assertIn("NAME\n    NoDefaults", help_screen)
        self.assertIn("SYNOPSIS\n    NoDefaults COMMAND", help_screen)
        self.assertNotIn("DESCRIPTION", help_screen)
        self.assertIn("COMMANDS\n    COMMAND is one of the following:", help_screen)
        self.assertIn("double", help_screen)
        self.assertIn("triple", help_screen)
        self.assertNotIn("NOTES", help_screen)

    def testHelpTextFunction(self):
        component = tc.NoDefaults().double
        help_screen = helptext.HelpText(component=component, trace=trace.FireTrace(component, name="double"))
        self.assertIn("NAME\n    double", help_screen)
        self.assertIn("SYNOPSIS\n    double COUNT", help_screen)
        self.assertNotIn("DESCRIPTION", help_screen)
        self.assertIn("POSITIONAL ARGUMENTS\n    COUNT", help_screen)
        self.assertIn("NOTES\n    You can also use flags syntax for POSITIONAL ARGUMENTS", help_screen)

    def testHelpTextFunctionWithDefaults(self):
        component = tc.WithDefaults().triple
        help_screen = helptext.HelpText(component=component, trace=trace.FireTrace(component, name="triple"))
        self.assertIn("NAME\n    triple", help_screen)
        self.assertIn("SYNOPSIS\n    triple <flags>", help_screen)
        self.assertNotIn("DESCRIPTION", help_screen)
        self.assertIn("FLAGS\n    -c, --count=COUNT\n        Default: 0", help_screen)
        self.assertNotIn("NOTES", help_screen)

    def testHelpTextFunctionWithLongDefaults(self):
        component = tc.WithDefaults().text
        help_screen = helptext.HelpText(component=component, trace=trace.FireTrace(component, name="text"))
        self.assertIn("NAME\n    text", help_screen)
        self.assertIn("SYNOPSIS\n    text <flags>", help_screen)
        self.assertNotIn("DESCRIPTION", help_screen)
        self.assertIn(
            "FLAGS\n    -s, --string=STRING\n"
            "        Default: '0001020304050607080910"
            "1112131415161718192021222324252627282...",
            help_screen,
        )
        self.assertNotIn("NOTES", help_screen)

    def testHelpTextFunctionWithKwargs(self):
        component = tc.fn_with_kwarg
        help_screen = helptext.HelpText(component=component, trace=trace.FireTrace(component, name="text"))
        self.assertIn("NAME\n    text", help_screen)
        self.assertIn("SYNOPSIS\n    text ARG1 ARG2 <flags>", help_screen)
        self.assertIn("DESCRIPTION\n    Function with kwarg", help_screen)
        self.assertIn(
            "FLAGS\n    --arg3\n        Description of arg3.\n    "
            "Additional undocumented flags may also be accepted.",
            help_screen,
        )

    def testHelpTextFunctionWithKwargsAndDefaults(self):
        component = tc.fn_with_kwarg_and_defaults
        help_screen = helptext.HelpText(component=component, trace=trace.FireTrace(component, name="text"))
        self.assertIn("NAME\n    text", help_screen)
        self.assertIn("SYNOPSIS\n    text ARG1 ARG2 <flags>", help_screen)
        self.assertIn("DESCRIPTION\n    Function with kwarg", help_screen)
        self.assertIn(
            "FLAGS\n    -o, --opt=OPT\n        Default: True\n"
            "    The following flags are also accepted."
            "\n    --arg3\n        Description of arg3.\n    "
            "Additional undocumented flags may also be accepted.",
            help_screen,
        )

    @testutils.skipIf(sys.version_info[0:2] < (3, 5), "Python < 3.5 does not support type hints.")
    def testHelpTextFunctionWithDefaultsAndTypes(self):
        component = tc.py3.WithDefaultsAndTypes().double  # pytype: disable=module-attr
        help_screen = helptext.HelpText(component=component, trace=trace.FireTrace(component, name="double"))
        self.assertIn("NAME\n    double", help_screen)
        self.assertIn("SYNOPSIS\n    double <flags>", help_screen)
        self.assertIn("DESCRIPTION", help_screen)
        self.assertIn("FLAGS\n    -c, --count=COUNT\n        Type: float\n        Default: 0", help_screen)
        self.assertNotIn("NOTES", help_screen)

    @testutils.skipIf(sys.version_info[0:2] < (3, 5), "Python < 3.5 does not support type hints.")
    def testHelpTextFunctionWithTypesAndDefaultNone(self):
        component = tc.py3.WithDefaultsAndTypes().get_int  # pytype: disable=module-attr
        help_screen = helptext.HelpText(component=component, trace=trace.FireTrace(component, name="get_int"))
        self.assertIn("NAME\n    get_int", help_screen)
        self.assertIn("SYNOPSIS\n    get_int <flags>", help_screen)
        self.assertNotIn("DESCRIPTION", help_screen)
        self.assertIn(
            "FLAGS\n    -v, --value=VALUE\n" "        Type: Optional[int]\n        Default: None", help_screen
        )
        self.assertNotIn("NOTES", help_screen)

    @testutils.skipIf(sys.version_info[0:2] < (3, 5), "Python < 3.5 does not support type hints.")
    def testHelpTextFunctionWithTypes(self):
        component = tc.py3.WithTypes().double  # pytype: disable=module-attr
        help_screen = helptext.HelpText(component=component, trace=trace.FireTrace(component, name="double"))
        self.assertIn("NAME\n    double", help_screen)
        self.assertIn("SYNOPSIS\n    double COUNT", help_screen)
        self.assertIn("DESCRIPTION", help_screen)
        self.assertIn("POSITIONAL ARGUMENTS\n    COUNT\n        Type: float", help_screen)
        self.assertIn("NOTES\n    You can also use flags syntax for POSITIONAL ARGUMENTS", help_screen)

    @testutils.skipIf(sys.version_info[0:2] < (3, 5), "Python < 3.5 does not support type hints.")
    def testHelpTextFunctionWithLongTypes(self):
        component = tc.py3.WithTypes().long_type  # pytype: disable=module-attr
        help_screen = helptext.HelpText(component=component, trace=trace.FireTrace(component, name="long_type"))
        self.assertIn("NAME\n    long_type", help_screen)
        self.assertIn("SYNOPSIS\n    long_type LONG_OBJ", help_screen)
        self.assertNotIn("DESCRIPTION", help_screen)
        # TODO(dbieber): Assert type is displayed correctly. Type displayed
        # differently in Travis vs in Google.
        # self.assertIn(
        #     'POSITIONAL ARGUMENTS\n    LONG_OBJ\n'
        #     '        Type: typing.Tuple[typing.Tuple['
        #     'typing.Tuple[typing.Tuple[typing.Tupl...',
        #     help_screen)
        self.assertIn("NOTES\n    You can also use flags syntax for POSITIONAL ARGUMENTS", help_screen)

    def testHelpTextFunctionWithBuiltin(self):
        component = "test".upper
        help_screen = helptext.HelpText(component=component, trace=trace.FireTrace(component, "upper"))
        self.assertIn("NAME\n    upper", help_screen)
        self.assertIn("SYNOPSIS\n    upper", help_screen)
        # We don't check description content here since the content is python
        # version dependent.
        self.assertIn("DESCRIPTION\n", help_screen)
        self.assertNotIn("NOTES", help_screen)

    def testHelpTextFunctionIntType(self):
        component = int
        help_screen = helptext.HelpText(component=component, trace=trace.FireTrace(component, "int"))
        self.assertIn("NAME\n    int", help_screen)
        self.assertIn("SYNOPSIS\n    int", help_screen)
        # We don't check description content here since the content is python
        # version dependent.
        self.assertIn("DESCRIPTION\n", help_screen)

    def testHelpTextEmptyList(self):
        component = []
        help_screen = helptext.HelpText(component=component, trace=trace.FireTrace(component, "list"))
        self.assertIn("NAME\n    list", help_screen)
        self.assertIn("SYNOPSIS\n    list COMMAND", help_screen)
        # TODO(zuhaochen): Change assertion after custom description is
        # implemented for list type.
        self.assertNotIn("DESCRIPTION", help_screen)
        # We don't check the listed commands either since the list API could
        # potentially change between Python versions.
        self.assertIn("COMMANDS\n    COMMAND is one of the following:\n", help_screen)

    def testHelpTextShortList(self):
        component = [10]
        help_screen = helptext.HelpText(component=component, trace=trace.FireTrace(component, "list"))
        self.assertIn("NAME\n    list", help_screen)
        self.assertIn("SYNOPSIS\n    list COMMAND", help_screen)
        # TODO(zuhaochen): Change assertion after custom description is
        # implemented for list type.
        self.assertNotIn("DESCRIPTION", help_screen)

        # We don't check the listed commands comprehensively since the list API
        # could potentially change between Python versions. Check a few
        # functions(command) that we're confident likely remain available.
        self.assertIn("COMMANDS\n    COMMAND is one of the following:\n", help_screen)
        self.assertIn("     append\n", help_screen)

    def testHelpTextInt(self):
        component = 7
        help_screen = helptext.HelpText(component=component, trace=trace.FireTrace(component, "7"))
        self.assertIn("NAME\n    7", help_screen)
        self.assertIn("SYNOPSIS\n    7 COMMAND | VALUE", help_screen)
        # TODO(zuhaochen): Change assertion after implementing custom
        # description for int.
        self.assertNotIn("DESCRIPTION", help_screen)
        self.assertIn("COMMANDS\n    COMMAND is one of the following:\n", help_screen)
        self.assertIn("VALUES\n    VALUE is one of the following:\n", help_screen)

    def testHelpTextNoInit(self):
        component = tc.OldStyleEmpty
        help_screen = helptext.HelpText(component=component, trace=trace.FireTrace(component, "OldStyleEmpty"))
        self.assertIn("NAME\n    OldStyleEmpty", help_screen)
        self.assertIn("SYNOPSIS\n    OldStyleEmpty", help_screen)

    @testutils.skipIf(six.PY2, "Python 2 does not support keyword-only arguments.")
    def testHelpTextKeywordOnlyArgumentsWithDefault(self):
        component = tc.py3.KeywordOnly.with_default  # pytype: disable=module-attr
        output = helptext.HelpText(component=component, trace=trace.FireTrace(component, "with_default"))
        self.assertIn("NAME\n    with_default", output)
        self.assertIn("FLAGS\n    -x, --x=X", output)

    @testutils.skipIf(six.PY2, "Python 2 does not support keyword-only arguments.")
    def testHelpTextKeywordOnlyArgumentsWithoutDefault(self):
        component = tc.py3.KeywordOnly.double  # pytype: disable=module-attr
        output = helptext.HelpText(component=component, trace=trace.FireTrace(component, "double"))
        self.assertIn("NAME\n    double", output)
        self.assertIn("FLAGS\n    -c, --count=COUNT (required)", output)

    @testutils.skipIf(six.PY2, "Python 2 does not support required name-only arguments.")
    def testHelpTextFunctionMixedDefaults(self):
        component = tc.py3.HelpTextComponent().identity
        t = trace.FireTrace(component, name="FunctionMixedDefaults")
        output = helptext.HelpText(component, trace=t)
        self.assertIn("NAME\n    FunctionMixedDefaults", output)
        self.assertIn("FunctionMixedDefaults <flags>", output)
        self.assertIn("--alpha=ALPHA (required)", output)
        self.assertIn("--beta=BETA\n        Default: '0'", output)

    def testHelpScreen(self):
        component = tc.ClassWithDocstring()
        t = trace.FireTrace(component, name="ClassWithDocstring")
        help_output = helptext.HelpText(component, t)
        expected_output = """
NAME
    ClassWithDocstring - Test class for testing help text output.

SYNOPSIS
    ClassWithDocstring COMMAND | VALUE

DESCRIPTION
    This is some detail description of this test class.

COMMANDS
    COMMAND is one of the following:

     print_msg
       Prints a message.

VALUES
    VALUE is one of the following:

     message
       The default message to print."""
        self.assertEqual(textwrap.dedent(expected_output).strip(), help_output.strip())

    def testHelpScreenForFunctionDocstringWithLineBreak(self):
        component = tc.ClassWithMultilineDocstring.example_generator
        t = trace.FireTrace(component, name="example_generator")
        help_output = helptext.HelpText(component, t)
        expected_output = """
    NAME
        example_generator - Generators have a ``Yields`` section instead of a ``Returns`` section.

    SYNOPSIS
        example_generator N

    DESCRIPTION
        Generators have a ``Yields`` section instead of a ``Returns`` section.

    POSITIONAL ARGUMENTS
        N
            The upper limit of the range to generate, from 0 to `n` - 1.

    NOTES
        You can also use flags syntax for POSITIONAL ARGUMENTS"""
        self.assertEqual(textwrap.dedent(expected_output).strip(), help_output.strip())

    def testHelpScreenForFunctionFunctionWithDefaultArgs(self):
        component = tc.WithDefaults().double
        t = trace.FireTrace(component, name="double")
        help_output = helptext.HelpText(component, t)
        expected_output = """
    NAME
        double - Returns the input multiplied by 2.

    SYNOPSIS
        double <flags>

    DESCRIPTION
        Returns the input multiplied by 2.

    FLAGS
        -c, --count=COUNT
            Default: 0
            Input number that you want to double."""
        self.assertEqual(textwrap.dedent(expected_output).strip(), help_output.strip())

    def testHelpTextUnderlineFlag(self):
        component = tc.WithDefaults().triple
        t = trace.FireTrace(component, name="triple")
        help_screen = helptext.HelpText(component, t)
        self.assertIn(formatting.Bold("NAME") + "\n    triple", help_screen)
        self.assertIn(formatting.Bold("SYNOPSIS") + "\n    triple <flags>", help_screen)
        self.assertIn(formatting.Bold("FLAGS") + "\n    -c, --" + formatting.Underline("count"), help_screen)

    def testHelpTextBoldCommandName(self):
        component = tc.ClassWithDocstring()
        t = trace.FireTrace(component, name="ClassWithDocstring")
        help_screen = helptext.HelpText(component, t)
        self.assertIn(formatting.Bold("NAME") + "\n    ClassWithDocstring", help_screen)
        self.assertIn(formatting.Bold("COMMANDS") + "\n", help_screen)
        self.assertIn(formatting.BoldUnderline("COMMAND") + " is one of the following:\n", help_screen)
        self.assertIn(formatting.Bold("print_msg") + "\n", help_screen)

    def testHelpTextObjectWithGroupAndValues(self):
        component = tc.TypedProperties()
        t = trace.FireTrace(component, name="TypedProperties")
        help_screen = helptext.HelpText(component=component, trace=t, verbose=True)
        print(help_screen)
        self.assertIn("GROUPS", help_screen)
        self.assertIn("GROUP is one of the following:", help_screen)
        self.assertIn("charlie\n       Class with functions that have default arguments.", help_screen)
        self.assertIn("VALUES", help_screen)
        self.assertIn("VALUE is one of the following:", help_screen)
        self.assertIn("alpha", help_screen)

    def testHelpTextNameSectionCommandWithSeparator(self):
        component = 9
        t = trace.FireTrace(component, name="int", separator="-")
        t.AddSeparator()
        help_screen = helptext.HelpText(component=component, trace=t, verbose=False)
        self.assertIn("int -", help_screen)
        self.assertNotIn("int - -", help_screen)

    def testHelpTextNameSectionCommandWithSeparatorVerbose(self):
        component = tc.WithDefaults().double
        t = trace.FireTrace(component, name="double", separator="-")
        t.AddSeparator()
        help_screen = helptext.HelpText(component=component, trace=t, verbose=True)
        self.assertIn("double -", help_screen)
        self.assertIn("double - -", help_screen)

    def testHelpTextMultipleKeywoardArgumentsWithShortArgs(self):
        component = tc.fn_with_multiple_defaults
        t = trace.FireTrace(component, name="shortargs")
        help_screen = helptext.HelpText(component, t)
        self.assertIn(formatting.Bold("NAME") + "\n    shortargs", help_screen)
        self.assertIn(formatting.Bold("SYNOPSIS") + "\n    shortargs <flags>", help_screen)
        self.assertIn(formatting.Bold("FLAGS") + "\n    -f, --first", help_screen)
        self.assertIn("\n    --last", help_screen)
        self.assertIn("\n    --late", help_screen)


class UsageTest(testutils.BaseTestCase):
    def testUsageOutput(self):
        component = tc.NoDefaults()
        t = trace.FireTrace(component, name="NoDefaults")
        usage_output = helptext.UsageText(component, trace=t, verbose=False)
        expected_output = """
    Usage: NoDefaults <command>
      available commands:    double | triple

    For detailed information on this command, run:
      NoDefaults --help"""

        self.assertEqual(usage_output, textwrap.dedent(expected_output).lstrip("\n"))

    def testUsageOutputVerbose(self):
        component = tc.NoDefaults()
        t = trace.FireTrace(component, name="NoDefaults")
        usage_output = helptext.UsageText(component, trace=t, verbose=True)
        expected_output = """
    Usage: NoDefaults <command>
      available commands:    double | triple

    For detailed information on this command, run:
      NoDefaults --help"""
        self.assertEqual(usage_output, textwrap.dedent(expected_output).lstrip("\n"))

    def testUsageOutputMethod(self):
        component = tc.NoDefaults().double
        t = trace.FireTrace(component, name="NoDefaults")
        t.AddAccessedProperty(component, "double", ["double"], None, None)
        usage_output = helptext.UsageText(component, trace=t, verbose=False)
        expected_output = """
    Usage: NoDefaults double COUNT

    For detailed information on this command, run:
      NoDefaults double --help"""
        self.assertEqual(usage_output, textwrap.dedent(expected_output).lstrip("\n"))

    def testUsageOutputFunctionWithHelp(self):
        component = tc.function_with_help
        t = trace.FireTrace(component, name="function_with_help")
        usage_output = helptext.UsageText(component, trace=t, verbose=False)
        expected_output = """
    Usage: function_with_help <flags>
      optional flags:        --help

    For detailed information on this command, run:
      function_with_help -- --help"""
        self.assertEqual(usage_output, textwrap.dedent(expected_output).lstrip("\n"))

    def testUsageOutputFunctionWithDocstring(self):
        component = tc.multiplier_with_docstring
        t = trace.FireTrace(component, name="multiplier_with_docstring")
        usage_output = helptext.UsageText(component, trace=t, verbose=False)
        expected_output = """
    Usage: multiplier_with_docstring NUM <flags>
      optional flags:        --rate

    For detailed information on this command, run:
      multiplier_with_docstring --help"""
        self.assertEqual(textwrap.dedent(expected_output).lstrip("\n"), usage_output)

    @testutils.skipIf(six.PY2, "Python 2 does not support required name-only arguments.")
    def testUsageOutputFunctionMixedDefaults(self):
        component = tc.py3.HelpTextComponent().identity
        t = trace.FireTrace(component, name="FunctionMixedDefaults")
        usage_output = helptext.UsageText(component, trace=t, verbose=False)
        expected_output = """
    Usage: FunctionMixedDefaults <flags>
      optional flags:        --beta
      required flags:        --alpha

    For detailed information on this command, run:
      FunctionMixedDefaults --help"""
        expected_output = textwrap.dedent(expected_output).lstrip("\n")
        self.assertEqual(expected_output, usage_output)

    def testUsageOutputCallable(self):
        # This is both a group and a command.
        component = tc.CallableWithKeywordArgument()
        t = trace.FireTrace(component, name="CallableWithKeywordArgument", separator="@")
        usage_output = helptext.UsageText(component, trace=t, verbose=False)
        expected_output = """
    Usage: CallableWithKeywordArgument <command> | <flags>
      available commands:    print_msg
      flags are accepted

    For detailed information on this command, run:
      CallableWithKeywordArgument -- --help"""
        self.assertEqual(textwrap.dedent(expected_output).lstrip("\n"), usage_output)

    def testUsageOutputConstructorWithParameter(self):
        component = tc.InstanceVars
        t = trace.FireTrace(component, name="InstanceVars")
        usage_output = helptext.UsageText(component, trace=t, verbose=False)
        expected_output = """
    Usage: InstanceVars --arg1=ARG1 --arg2=ARG2

    For detailed information on this command, run:
      InstanceVars --help"""
        self.assertEqual(textwrap.dedent(expected_output).lstrip("\n"), usage_output)

    def testUsageOutputConstructorWithParameterVerbose(self):
        component = tc.InstanceVars
        t = trace.FireTrace(component, name="InstanceVars")
        usage_output = helptext.UsageText(component, trace=t, verbose=True)
        expected_output = """
    Usage: InstanceVars <command> | --arg1=ARG1 --arg2=ARG2
      available commands:    run

    For detailed information on this command, run:
      InstanceVars --help"""
        self.assertEqual(textwrap.dedent(expected_output).lstrip("\n"), usage_output)

    def testUsageOutputEmptyDict(self):
        component = {}
        t = trace.FireTrace(component, name="EmptyDict")
        usage_output = helptext.UsageText(component, trace=t, verbose=True)
        expected_output = """
    Usage: EmptyDict

    For detailed information on this command, run:
      EmptyDict --help"""
        self.assertEqual(textwrap.dedent(expected_output).lstrip("\n"), usage_output)

    def testUsageOutputNone(self):
        component = None
        t = trace.FireTrace(component, name="None")
        usage_output = helptext.UsageText(component, trace=t, verbose=True)
        expected_output = """
    Usage: None

    For detailed information on this command, run:
      None --help"""
        self.assertEqual(textwrap.dedent(expected_output).lstrip("\n"), usage_output)

    def testInitRequiresFlagSyntaxSubclassNamedTuple(self):
        component = tc.SubPoint
        t = trace.FireTrace(component, name="SubPoint")
        usage_output = helptext.UsageText(component, trace=t, verbose=False)
        expected_output = "Usage: SubPoint --x=X --y=Y"
        self.assertIn(expected_output, usage_output)


if __name__ == "__main__":
    testutils.main()



# Copyright (C) 2018 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the completion module."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from afire import completion
from afire import test_components as tc
from afire import testutils


class TabCompletionTest(testutils.BaseTestCase):
    def testCompletionBashScript(self):
        # A sanity check test to make sure the bash completion script satisfies
        # some basic assumptions.
        commands = [
            ["run"],
            ["halt"],
            ["halt", "--now"],
        ]
        script = completion._BashScript(name="command", commands=commands)  # pylint: disable=protected-access
        self.assertIn("command", script)
        self.assertIn("halt", script)

        assert_template = "{command})"
        for last_command in ["command", "halt"]:
            self.assertIn(assert_template.format(command=last_command), script)

    def testCompletionFishScript(self):
        # A sanity check test to make sure the fish completion script satisfies
        # some basic assumptions.
        commands = [
            ["run"],
            ["halt"],
            ["halt", "--now"],
        ]
        script = completion._FishScript(name="command", commands=commands)  # pylint: disable=protected-access
        self.assertIn("command", script)
        self.assertIn("halt", script)
        self.assertIn("-l now", script)

    def testFnCompletions(self):
        def example(one, two, three):
            return one, two, three

        completions = completion.Completions(example)
        self.assertIn("--one", completions)
        self.assertIn("--two", completions)
        self.assertIn("--three", completions)

    def testListCompletions(self):
        completions = completion.Completions(["red", "green", "blue"])
        self.assertIn("0", completions)
        self.assertIn("1", completions)
        self.assertIn("2", completions)
        self.assertNotIn("3", completions)

    def testDictCompletions(self):
        colors = {
            "red": "green",
            "blue": "yellow",
            "_rainbow": True,
        }
        completions = completion.Completions(colors)
        self.assertIn("red", completions)
        self.assertIn("blue", completions)
        self.assertNotIn("green", completions)
        self.assertNotIn("yellow", completions)
        self.assertNotIn("_rainbow", completions)
        self.assertNotIn("True", completions)
        self.assertNotIn(True, completions)

    def testDictCompletionsVerbose(self):
        colors = {
            "red": "green",
            "blue": "yellow",
            "_rainbow": True,
        }
        completions = completion.Completions(colors, verbose=True)
        self.assertIn("red", completions)
        self.assertIn("blue", completions)
        self.assertNotIn("green", completions)
        self.assertNotIn("yellow", completions)
        self.assertIn("_rainbow", completions)
        self.assertNotIn("True", completions)
        self.assertNotIn(True, completions)

    def testDeepDictCompletions(self):
        deepdict = {"level1": {"level2": {"level3": {"level4": {}}}}}
        completions = completion.Completions(deepdict)
        self.assertIn("level1", completions)
        self.assertNotIn("level2", completions)

    def testDeepDictScript(self):
        deepdict = {"level1": {"level2": {"level3": {"level4": {}}}}}
        script = completion.Script("deepdict", deepdict)
        self.assertIn("level1", script)
        self.assertIn("level2", script)
        self.assertIn("level3", script)
        self.assertNotIn("level4", script)  # The default depth is 3.

    def testFnScript(self):
        script = completion.Script("identity", tc.identity)
        self.assertIn("--arg1", script)
        self.assertIn("--arg2", script)
        self.assertIn("--arg3", script)
        self.assertIn("--arg4", script)

    def testClassScript(self):
        script = completion.Script("", tc.MixedDefaults)
        self.assertIn("ten", script)
        self.assertIn("sum", script)
        self.assertIn("identity", script)
        self.assertIn("--alpha", script)
        self.assertIn("--beta", script)

    def testDeepDictFishScript(self):
        deepdict = {"level1": {"level2": {"level3": {"level4": {}}}}}
        script = completion.Script("deepdict", deepdict, shell="fish")
        self.assertIn("level1", script)
        self.assertIn("level2", script)
        self.assertIn("level3", script)
        self.assertNotIn("level4", script)  # The default depth is 3.

    def testFnFishScript(self):
        script = completion.Script("identity", tc.identity, shell="fish")
        self.assertIn("arg1", script)
        self.assertIn("arg2", script)
        self.assertIn("arg3", script)
        self.assertIn("arg4", script)

    def testClassFishScript(self):
        script = completion.Script("", tc.MixedDefaults, shell="fish")
        self.assertIn("ten", script)
        self.assertIn("sum", script)
        self.assertIn("identity", script)
        self.assertIn("alpha", script)
        self.assertIn("beta", script)

    def testNonStringDictCompletions(self):
        completions = completion.Completions(
            {
                10: "green",
                3.14: "yellow",
                ("t1", "t2"): "pink",
            }
        )
        self.assertIn("10", completions)
        self.assertIn("3.14", completions)
        self.assertIn("('t1', 't2')", completions)
        self.assertNotIn("green", completions)
        self.assertNotIn("yellow", completions)
        self.assertNotIn("pink", completions)

    def testGeneratorCompletions(self):
        def generator():
            x = 0
            while True:
                yield x
                x += 1

        completions = completion.Completions(generator())
        self.assertEqual(completions, [])

    def testClassCompletions(self):
        completions = completion.Completions(tc.NoDefaults)
        self.assertEqual(completions, [])

    def testObjectCompletions(self):
        completions = completion.Completions(tc.NoDefaults())
        self.assertIn("double", completions)
        self.assertIn("triple", completions)

    def testMethodCompletions(self):
        completions = completion.Completions(tc.NoDefaults().double)
        self.assertNotIn("--self", completions)
        self.assertIn("--count", completions)


if __name__ == "__main__":
    testutils.main()



# Copyright (C) 2018 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test using Fire via `python -m fire`."""

import os
import tempfile

from afire import __main__
from afire import testutils


class MainModuleTest(testutils.BaseTestCase):
    """Tests to verify the behavior of __main__ (python -m fire)."""

    def testNameSetting(self):
        # Confirm one of the usage lines has the gettempdir member.
        with self.assertOutputMatches("gettempdir"):
            __main__.main(["__main__.py", "tempfile"])

    def testArgPassing(self):
        expected = os.path.join("part1", "part2", "part3")
        with self.assertOutputMatches("%s\n" % expected):
            __main__.main(["__main__.py", "os.path", "join", "part1", "part2", "part3"])
        with self.assertOutputMatches("%s\n" % expected):
            __main__.main(["__main__.py", "os", "path", "-", "join", "part1", "part2", "part3"])


class MainModuleFileTest(testutils.BaseTestCase):
    """Tests to verify correct import behavior for file executables."""

    def setUp(self):
        super(MainModuleFileTest, self).setUp()
        self.file = tempfile.NamedTemporaryFile(suffix=".py")  # pylint: disable=consider-using-with
        self.file.write(b"class Foo:\n  def double(self, n):\n    return 2 * n\n")
        self.file.flush()

        self.file2 = tempfile.NamedTemporaryFile()  # pylint: disable=consider-using-with

    def testFileNameFire(self):
        # Confirm that the file is correctly imported and doubles the number.
        with self.assertOutputMatches("4"):
            __main__.main(["__main__.py", self.file.name, "Foo", "double", "--n", "2"])

    def testFileNameFailure(self):
        # Confirm that an existing file without a .py suffix raises a ValueError.
        with self.assertRaises(ValueError):
            __main__.main(["__main__.py", self.file2.name, "Foo", "double", "--n", "2"])

    def testFileNameModuleDuplication(self):
        # Confirm that a file that masks a module still loads the module.
        with self.assertOutputMatches("gettempdir"):
            dirname = os.path.dirname(self.file.name)
            with testutils.ChangeDirectory(dirname):
                with open("tempfile", "w"):
                    __main__.main(
                        [
                            "__main__.py",
                            "tempfile",
                        ]
                    )

                os.remove("tempfile")

    def testFileNameModuleFileFailure(self):
        # Confirm that an invalid file that masks a non-existent module fails.
        with self.assertRaisesRegex(
            ValueError, r"Fire can only be called on \.py files\."
        ):  # pylint: disable=line-too-long,  # pytype: disable=attribute-error
            dirname = os.path.dirname(self.file.name)
            with testutils.ChangeDirectory(dirname):
                with open("foobar", "w"):
                    __main__.main(
                        [
                            "__main__.py",
                            "foobar",
                        ]
                    )

                os.remove("foobar")


if __name__ == "__main__":
    testutils.main()



# Copyright (C) 2018 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Types of values."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import inspect

from afire import inspectutils
import six


VALUE_TYPES = (
    bool,
    six.string_types,
    six.integer_types,
    float,
    complex,
    type(Ellipsis),
    type(None),
    type(NotImplemented),
)


def IsGroup(component):
    # TODO(dbieber): Check if there are any subcomponents.
    return not IsCommand(component) and not IsValue(component)


def IsCommand(component):
    return inspect.isroutine(component) or inspect.isclass(component)


def IsValue(component):
    return isinstance(component, VALUE_TYPES) or HasCustomStr(component)


def IsSimpleGroup(component):
    """If a group is simple enough, then we treat it as a value in PrintResult.

    Only if a group contains all value types do we consider it simple enough to
    print as a value.

    Args:
      component: The group to check for value-group status.
    Returns:
      A boolean indicating if the group should be treated as a value for printing
      purposes.
    """
    assert isinstance(component, dict)
    for unused_key, value in component.items():
        if not IsValue(value) and not isinstance(value, (list, dict)):
            return False
    return True


def HasCustomStr(component):
    """Determines if a component has a custom __str__ method.

    Uses inspect.classify_class_attrs to determine the origin of the object's
    __str__ method, if one is present. If it defined by `object` itself, then
    it is not considered custom. Otherwise it is. This means that the __str__
    methods of primitives like ints and floats are considered custom.

    Objects with custom __str__ methods are treated as values and can be
    serialized in places where more complex objects would have their help screen
    shown instead.

    Args:
      component: The object to check for a custom __str__ method.
    Returns:
      Whether `component` has a custom __str__ method.
    """
    if hasattr(component, "__str__"):
        class_attrs = inspectutils.GetClassAttrsDict(type(component)) or {}
        str_attr = class_attrs.get("__str__")
        if str_attr and str_attr.defining_class is not object:
            return True
    return False



# Copyright (C) 2018 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for Python Fire's tests."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import contextlib
import os
import re
import sys
import unittest

from afire import core
from afire import trace

import mock
import six


class BaseTestCase(unittest.TestCase):
    """Shared test case for Python Fire tests."""

    @contextlib.contextmanager
    def assertOutputMatches(self, stdout=".*", stderr=".*", capture=True):
        """Asserts that the context generates stdout and stderr matching regexps.

        Note: If wrapped code raises an exception, stdout and stderr will not be
          checked.

        Args:
          stdout: (str) regexp to match against stdout (None will check no stdout)
          stderr: (str) regexp to match against stderr (None will check no stderr)
          capture: (bool, default True) do not bubble up stdout or stderr

        Yields:
          Yields to the wrapped context.
        """
        stdout_fp = six.StringIO()
        stderr_fp = six.StringIO()
        try:
            with mock.patch.object(sys, "stdout", stdout_fp):
                with mock.patch.object(sys, "stderr", stderr_fp):
                    yield
        finally:
            if not capture:
                sys.stdout.write(stdout_fp.getvalue())
                sys.stderr.write(stderr_fp.getvalue())

        for name, regexp, fp in [("stdout", stdout, stdout_fp), ("stderr", stderr, stderr_fp)]:
            value = fp.getvalue()
            if regexp is None:
                if value:
                    raise AssertionError("%s: Expected no output. Got: %r" % (name, value))
            else:
                if not re.search(regexp, value, re.DOTALL | re.MULTILINE):
                    raise AssertionError("%s: Expected %r to match %r" % (name, value, regexp))

    def assertRaisesRegex(self, *args, **kwargs):  # pylint: disable=arguments-differ
        if sys.version_info.major == 2:
            return super(BaseTestCase, self).assertRaisesRegexp(*args, **kwargs)  # pylint: disable=deprecated-method
        else:
            return super(BaseTestCase, self).assertRaisesRegex(*args, **kwargs)  # pylint: disable=no-member

    @contextlib.contextmanager
    def assertRaisesFireExit(self, code, regexp=".*"):
        """Asserts that a FireExit error is raised in the context.

        Allows tests to check that Fire's wrapper around SystemExit is raised
        and that a regexp is matched in the output.

        Args:
          code: The status code that the FireExit should contain.
          regexp: stdout must match this regex.

        Yields:
          Yields to the wrapped context.
        """
        with self.assertOutputMatches(stderr=regexp):
            with self.assertRaises(core.FireExit):
                try:
                    yield
                except core.FireExit as exc:
                    if exc.code != code:
                        raise AssertionError("Incorrect exit code: %r != %r" % (exc.code, code))
                    self.assertIsInstance(exc.trace, trace.FireTrace)
                    raise


@contextlib.contextmanager
def ChangeDirectory(directory):
    """Context manager to mock a directory change and revert on exit."""
    cwdir = os.getcwd()
    os.chdir(directory)

    try:
        yield directory
    finally:
        os.chdir(cwdir)


# pylint: disable=invalid-name
main = unittest.main
skip = unittest.skip
skipIf = unittest.skipIf
# pylint: enable=invalid-name
